from typing import Iterable


def assert_true(condition):
    if condition:
        return True
    else:
        raise AssertionError("condition is False")

def assert_false(condition):
    if not condition:
        return True
    else:
        raise AssertionError("condition is True")

def assert_equal(a, b):
    if a == b:
        return True
    else:
        raise AssertionError(f"{a} != {b}")

def assert_not_equal(a, b):
    if a != b:
        return True
    else:
        raise AssertionError(f"{a} == {b}")

def assert_is(a, b):
    if a is b:
        return True
    else:
        raise AssertionError(f"{a} is not {b}")

def assert_greater(a, b):
    if a > b:
        return True
    else:
        raise AssertionError(f"{a} <= {b}")

def assert_greater_equal(a, b):
    if a >= b:
        return True
    else:
        raise AssertionError(f"{a} < {b}")

def assert_less(a, b):
    if a < b:
        return True
    else:
        raise AssertionError(f"{a} >= {b}")

def assert_less_equal(a, b):
    if a <= b:
        return True
    else:
        raise AssertionError(f"{a} > {b}")

def assert_close(a, b, epsilon=1e-6):
    if abs(a - b) < epsilon:
        return True
    else:
        raise AssertionError(f"{a} != {b}")

def assert_type(obj, type):
    if not isinstance(obj, type):
        if isinstance(type, Iterable):
            if len(type) > 0:
                type_str = type[-1].__name__
            if len(type) > 1:
                type_str = type[-2].__name__ + " or " + type_str
            for i in range(3, len(type) + 1):
                type_str = type[-i].__name__ + ", " + type_str
        else:
            type_str = type.__name__
        raise AssertionError(f"object is not an instance of {type_str}.")
    return True

def assert_types(objects, types):
    for index, obj, type in enumerate(zip(objects, types)):
        try:
            assert_type(obj, type)
        except AssertionError as e:
            if isinstance(type, Iterable):
                if len(type) > 0:
                    type_str = type[-1].__name__
                if len(type) > 1:
                    type_str = type[-2].__name__ + " or " + type_str
                for i in range(3, len(type) + 1):
                    type_str = type[-i].__name__ + ", " + type_str
            else:
                type_str = type.__name__
            raise AssertionError(
                f"object number {index + 1} is not an instance of {type_str}."
            ) from e
    return True

del Iterable


from multiprocessing import Manager
import multiprocessing
import sys
from time import sleep, time
from datetime import datetime

class ProgressBar:
    """
    class for easily creating a progress bar
    """
    def __init__(
        self,
        max_value=100,
        width=20,
        log_interval=0.1,
        display_percentage=True,
        display_time_passed=True,
        display_time_remaining=True,
        draw_function=None,
    ):
        self._progress = Manager().Value("progress", 0)
        self._cancel = Manager().Value("cancel", False)

        self.max_value = max_value

        if draw_function is None:
            draw_function = ProgressBar.draw_bar

        self.process = multiprocessing.Process(
            target=ProgressBar.run,
            args=(
                self._progress,
                self._cancel,
                width,
                max_value,
                log_interval,
                display_percentage,
                display_time_passed,
                display_time_remaining,
                draw_function,
            ),
        )

        self.process.start()

    @staticmethod
    def run(
        progress,
        cancel,
        width,
        max_value,
        log_interval,
        display_percentage,
        display_time_passed,
        display_time_remaining,
        draw_function,
    ):
        start_time = datetime.now()
        while not cancel.value:
            if progress.value >= max_value:
                cancel.value = True

            bar_string = draw_function(progress.value, width, max_value)
            if display_percentage:
                bar_string += (
                    " {:.2f}%".format(progress.value / max_value * 100)
                )
            time_delta = datetime.now() - start_time
            if display_time_passed:
                bar_string += " | Time passed: " + str(time_delta)
            if display_time_remaining:
                remaining_time_estimate = (
                    time_delta * (max_value / (progress.value if progress.value != 0 else 0.001) - 1)
                ) 
                bar_string += " | Time remaining: " + str(remaining_time_estimate)            
            while len(bar_string) < width + 20:
                bar_string += " "
            sys.stdout.flush()
            sys.stdout.write("\r" + bar_string)
            sleep(log_interval)
        sys.stdout.write("\n")
    
    def update(self, value):
        self._progress.value = value

    def cancel(self):
        self._cancel.value = True
        self.process.join()
    
    def finish(self):
        self._progress.value = self.max_value
        self.process.join()

    def await_done(self):
        self.process.join()

    @staticmethod
    def draw_bar(progress, width, max_value):
        return_str = "】"
        progress_width = progress / max_value * width
        return_str += "█" * int(progress_width)#
        rest = progress_width - int(progress_width)
        if 0 == rest and progress_width < width:
            return_str += " "
        elif 0 < rest < 0.125:
            return_str += "▏"
        elif 0.125 <= rest < 0.25:
            return_str += "▎"
        elif 0.25 <= rest < 0.375:
            return_str += "▍"
        elif 0.375 <= rest < 0.5:
            return_str += "▌"
        elif 0.5 <= rest < 0.625:
            return_str += "▋"
        elif 0.625 <= rest < 0.75:
            return_str += "▊"
        elif 0.75 <= rest < 0.875:
            return_str += "▉"
        elif 0.875 <= rest < 1:
            return_str += "█"
            
        return_str += " " * int(width - progress_width)
        return_str += "【"
        return return_str

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.finish()
        else:
            self.cancel()

    @property
    def progress(self):
        return self._progress.value

    @progress.setter
    def progress(self, value):
        self._progress.value = value


from typing import Generic, Tuple, Type, TypeVar
from ..mutable._mutable import Mutable
from ._operators import *
from inspect import signature


T = TypeVar('T')


class Link(Mutable[T], Generic[T]):
    def __init__(self, operator: Type[Operator], *values: "Tuple[T, ...]"):
        assert issubclass(operator, Operator)
        self.operator = operator
        self.inputs = list(values)

        values_given = len(values)
        values_expected = signature(operator._eval).parameters.__len__()
        if values_given != values_expected:
            for attr in signature(operator._eval).parameters.values():
                # If an unpack operation is used, the given arguments will not exceed the expected arguments.
                if str(attr).startswith("*") and not str(attr).startswith("**"):
                    break
            else:
                raise ValueError(f"wrong number of values for {operator.__name__} (expected {values_expected}, got {values_given})")
    
    @property
    def value(self) -> T:
        inputs = (
            value.value 
            if isinstance(value, Mutable) 
            else value
            for value 
            in self.inputs
        )

        return self.operator(*inputs)
    
    @value.setter
    def value(self, value):
        # Modify the first value using the reverse operation
        try:
            value1 = self.operator.reverse(value, *self.inputs[1:])
            if value1 is not None:
                if isinstance(self.inputs[0], Mutable):
                    self.inputs[0].__set__(self, value1)
                else:
                    self.inputs[0] = value1
            else:
                raise ValueError(f"value cannot be set: {self.operator.__name__}({self.inputs[0]}, <Any>) != {value}")
        except NotImplementedError as e:
            print(e.with_traceback())
            print("cannot set value for link without reverse operation")
    
    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(map(repr, self.inputs))})"
    
    def __add__(self, other):
        return Add(self, other)
    
    def __radd__(self, other):
        return Add(other, self)
    
    def __sub__(self, other):
        return Sub(self, other)

    def __rsub__(self, other):
        return RSub(self, other)
    
    def __mul__(self, other):
        return Mul(self, other)
    
    def __rmul__(self, other):
        return Mul(other, self)
    
    def __truediv__(self, other):
        return Div(self, other)
    
    def __rtruediv__(self, other):
        return RDiv(self, other)
    
    def __pow__(self, other):
        return Pow(self, other)

    def __rpow__(self, other):
        return RPow(self, other)
    
    def __mod__(self, other):
        return Mod(self, other)
    
    def __rmod__(self, other):
        return Mod(other, self)

    def __abs__(self):
        return Abs(self)
    
    def __neg__(self):
        return Not(self)

    def __invert__(self):
        return Not(self)
    
    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Not(Eq(self, other))
    
    def __gt__(self, other):
        return Gt(self, other)
    
    def __ge__(self, other):
        return Ge(self, other)
    
    def __lt__(self, other):
        return Lt(self, other)
    
    def __le__(self, other):
        return Le(self, other)
    
    def __and__(self, other):
        return And(self, other)
    
    def __rand__(self, other):
        return And(other, self)
    
    def __or__(self, other):
        return Or(self, other)
    
    def __ror__(self, other):
        return Or(other, self)
    
    def __xor__(self, other):
        return Xor(self, other)
    
    def __rxor__(self, other):
        return Xor(other, self)

    
    def __iadd__(self, other):
        other_value = other.value if isinstance(other, Link) else other
        self.value += other_value
        return self


class Value(Link[T]):
    def __init__(self, value: T):
        super().__init__(ValueOperator, value)

class Eq(Link[bool]):
    def __init__(self, a: Link, b: Link):
        super().__init__(EqualOperator, a, b)

class Gt(Link[bool]):
    def __init__(self, a: Link, b: Link):
        super().__init__(GreaterOperator, a, b)

class Ge(Link[bool]):
    def __init__(self, a: Link, b: Link):
        super().__init__(GreaterEqualOperator, a, b)

class Lt(Link[bool]):
    def __init__(self, a: Link, b: Link):
        super().__init__(LessOperator, a, b)

class Le(Link[bool]):
    def __init__(self, a: Link, b: Link):
        super().__init__(LessEqualOperator, a, b)


class Add(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(AdditionOperator, a, b)

class Sub(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(SubtractionOperator, a, b)

class RSub(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsSubtractionOperator, a, b)

class Mul(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(MultiplicationOperator, a, b)

class Div(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(DivisionOperator, a, b)

class RDiv(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsDivisionOperator, a, b)

class Pow(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(PowerOperator, a, b)

class RPow(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsPowerOperator, a, b)

class Root(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(RootOperator, a, b)

class RRoot(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsRootOperator, a, b)

class Mod(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(ModuloOperator, a, b)

class Abs(Link[T]):
    def __init__(self, a: "Link[T] | T"):
        super().__init__(AbsoluteOperator, a)


class And(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(AndOperator, a, b)

class Or(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(OrOperator, a, b)

class Xor(Link[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(XorOperator, a, b)

class Not(Link[T]):
    def __init__(self, a: "Link[T] | T"):
        super().__init__(NotOperator, a)


"""
## the little things
a library full of small utilities for you to use in your code
"""

from .progress_bar import *
from . import auto_reload
from .mutable import *
from .variables import *
from .linked_values import *
from .testing import *
from .debug import *
from .extended_list import EList
from .constants import *
from .testing import *
from . import to_string

import atexit

def reset_styles():
    import sys

    try:
        if not translate_color_codes(Log._full_str, console=False).endswith("\n"):
            sys.stdout.write("\n")
    except SyntaxError:
        pass
    sys.stdout.write(translate_color_codes("[style: reset_all]\r"))
    sys.stdout.flush()
    
    Log.close_file()

atexit.register(reset_styles)
del atexit



import os
import sys
from ._coloring import translate_color_codes
from ..constants import UNDEFINED
from ..files import load_file


class Log:
    print = True
    _file = None
    sep = " "
    end = "\n"

    _full_str = ""

    def __new__(cls, *values, sep=UNDEFINED, end=UNDEFINED, print_=UNDEFINED, file_path=UNDEFINED):
        file = load_file(file_path, True, default=None) or cls._file
        print_ = cls.print if print_ is UNDEFINED else print_
        sep = cls.sep if sep is UNDEFINED else sep
        end = cls.end if end is UNDEFINED else end

        if print_:
            value_list = list(values)
            for i in range(len(value_list)):
                if isinstance(value_list[i], str):
                    value_list[i] = translate_color_codes(value_list[i], console=True)
            sys.stdout.flush()
            sys.stdout.write(sep.join(str(value) for value in value_list) + end)
    
        value_list = list(values)
        for i in range(len(value_list)):
            if isinstance(value_list[i], str):
                value_list[i] = translate_color_codes(value_list[i], console=False)
        write_string = sep.join(str(value) for value in value_list) + end

        cls._full_str += write_string

        if file is not None:
            file.write(write_string)
    
    @classmethod
    def load_file(cls, file_path, create_if_nonexistant=True):
        cls.close_file()
        cls._file = load_file(file_path, create_if_nonexistant)
    
    @classmethod
    def close_file(cls):
        if cls._file is not None:
            cls._file.close()
            cls._file = None


from typing import List, TypeVar, Type
from ..extended_list import EList

T = TypeVar('T')

def get_all_subclasses(cls: Type[T]) -> List[Type[T]]:
    """
    get all subclasses and subclasses of subclasses of a class.
    """

    all_subclasses = EList([])

    for subclass in cls.__subclasses__():
        all_subclasses.merge([subclass])
        all_subclasses.merge(get_all_subclasses(subclass))

    return all_subclasses


from colorama import Fore, Style, Back
from .. import to_string


def translate_color_codes(string: str, console=True):
    # Parse for [<Attribute>: <Value>, ...] patterns. Replaces colorama as logging colorama characters
    # to a file results in the escape characters being shown in unwanted ways.

    i = 0

    legal_attributes = {
        "fore": Fore,
        "foreground": Fore,
        "text": Fore,
        "color": Fore,
        "style": Style,
        "back": Back,
        "background": Back,
        "fill": Back,
    }

    while i < len(string):
        if string[i] == "[":
            i += 1
            if string[i] == "\\":
                string = string[:i] + string[i+1:]
            else:
                expression_start = i
                attribute_value_pairs = {}
                while True:
                    attribute = ""
                    try:
                        while string[i] != ":":
                            if string[i] == "]":
                                raise SyntaxError(f"In string \"{string}\": String colorization requires attribute and value. Syntax: [<Attribute>: <value>, ...]")
                            attribute += string[i]
                            i += 1

                        i += 1

                        while string[i] == " ":
                            i += 1
                        
                        value = ""
                        while string[i] != "]" and string[i] != " " and string[i] != ",":
                            value += string[i]
                            i += 1
                        
                        if value == "":
                            raise SyntaxError(f"In string \"{string}\": No value given.")
                        
                        attribute_value_pairs[attribute.lower()] = value

                        while string[i] == " ":
                            i += 1
                        
                        if string[i] != ",":
                            break
                        i += 1
                        while string[i] == " ":
                            i += 1
                    except IndexError as e:
                        raise SyntaxError(f"In string \"{string}\": \"[\" was not closed.") from e

                colorization_string = ""

                if console:
                    for key in attribute_value_pairs:
                        if key not in legal_attributes:
                            valid_attributes = to_string.list([f'"{attr.title()}"' for attr in list(legal_attributes)])
                            raise AttributeError(f"Attribute \"{key}\" not found. Valid attributes are {valid_attributes}")
                        try:
                            colorization_string += getattr(legal_attributes[key], attribute_value_pairs[key].upper())
                        except AttributeError as e:
                            valid_values = to_string.list([f'"{value.lower()}"' for value in dir(legal_attributes[key]) if not value.startswith("_")])
                            raise ValueError(f"Invalid value for {key.title()}: {attribute_value_pairs[key]}. Valid values are {valid_values}") from e

                string = string[:expression_start-1] + colorization_string + string[i+1:]
                i = expression_start + len(colorization_string) - 2
                
        i += 1
        
    return string


class Operator:
    # TODO: Add proper min and max values
    def __init__(self):
        self.min_values = {}
        self.max_values = {}
    
    def __new__(cls, *args, **kwargs):
        return cls._eval(*args, **kwargs)
    
    @classmethod
    def _eval(cls, *values):
        raise NotImplementedError(f"operation not implemented for {cls.__name__}")

    # Reverse operation.
    # For this:
    # c = Operation().call(a, b)
    # This should be true:
    # a = Operation().reverse(b, c)  
    @classmethod
    def _eval_reverse(cls, result, *values):
        raise NotImplementedError(f"reverse operation not implemented for {cls.__name__}")

    @classmethod
    def reverse(cls, result, *values):
        return cls._eval_reverse(result, *values)

    @classmethod
    def has_reverse(cls):
        try:
            cls._eval_reverse(*(1 for _ in range(len(cls._eval_reverse.__code__.co_varnames) - 1)))
            return True
        except NotImplementedError:
            return False        

class ValueOperator(Operator):
    @classmethod
    def _eval(cls, a):
        return a
    
    @classmethod
    def _eval_reverse(cls, b):
        return b

class EqualOperator(Operator):
    @classmethod
    def _eval(cls, a, b):
        return a == b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class GreaterOperator(Operator):
    @classmethod
    def _eval(cls, a, b):
        return a > b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class GreaterEqualOperator(Operator):
    @classmethod
    def _eval(cls, a, b):
        return a >= b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None

class LessOperator(Operator):
    @classmethod
    def _eval(cls, a, b):
        return a < b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None
            
class LessEqualOperator(Operator):
    @classmethod
    def _eval(cls, a, b):
        return a <= b
    
    @classmethod
    def _eval_reverse(cls, b, c):
        if c:
            return b
        else:
            return None


class NumberOperator(Operator):
    @classmethod
    def _eval(cls, *values):
        return super()._eval(cls)

    @classmethod
    def _eval_reverse(cls, c, b):
        return super()._eval_reverse(c, b)

class AdditionOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a + b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c - b

class SubtractionOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a - b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c + b

# All positional operators that have a reverse operation need a
# corresponding backwards operator.
class BackwardsSubtractionOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        return b - a
    
    @classmethod
    def _eval_reverse(cls, c, a):
        return c + a

class MultiplicationOperator(NumberOperator):     
    @classmethod
    def _eval(cls, a, b):
        return a * b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == 0:
            raise ZeroDivisionError("tried to divide by zero in MultiplicationOperator.reverse")
        return c / b

class DivisionOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        if b == 0:
            raise ZeroDivisionError("tried to divide by zero in DivisionOperator")
        return a / b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c * b

class BackwardsDivisionOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        if a == 0:
            raise ZeroDivisionError("tried to divide by zero in BackwardsDivide")
        return b / a
    
    @classmethod
    def _eval_reverse(cls, c, a):
        return c * a

class PowerOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a ** b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == 0:
            raise ArithmeticError("tried to take zeroth root in PowerOperator.reverse")
        return c ** (1 / b)

class BackwardsPowerOperator(NumberOperator):
    @classmethod
    def _eval(cls, a, b):
        return b ** a
    
    @classmethod
    def _eval_reverse(cls, c, a):
        if a == 0:
            raise ArithmeticError("tried to take zeroth root in BackwardsPowerOperator.reverse")
        return c ** (1 / a)

class RootOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        if b == 0:
            raise ArithmeticError("tried to take zeroth root in RootOperator")
        return a ** (1 / b)
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c ** b

class BackwardsRootOperator(NumberOperator):
    @classmethod
    def _eval(cls, a, b):
        if a == 0:
            raise ArithmeticError("tried to take zeroth root in BackwardsRootOperator")
        return b ** (1 / a)
    
    @classmethod
    def _eval_reverse(cls, c, a):
        return c ** a

class ModuloOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a % b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        raise NotImplementedError("ModuloOperator does not have a reverse operation")

class AbsoluteOperator(NumberOperator):    
    @classmethod
    def _eval(cls, a):
        return abs(a)
    
    @classmethod
    def _eval_reverse(cls, c, b):
        raise NotImplementedError("AbsOperator does not have a reverse operation")

class BooleanOperator(Operator):
    @classmethod
    def _eval(cls, *values):
        return super()._eval(cls)

    @classmethod
    def _eval_reverse(cls, result, *values):
        return super()._eval_reverse(result, *values)
        
class AndOperator(BooleanOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a and b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == False:
            return None
        else:
            return c

class OrOperator(BooleanOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a or b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        if b == True:
            return None
        else:
            return c

class XorOperator(BooleanOperator):    
    @classmethod
    def _eval(cls, a, b):
        return a ^ b
    
    @classmethod
    def _eval_reverse(cls, c, b):
        return c ^ b

class NotOperator(BooleanOperator):    
    @classmethod
    def _eval(cls, a):
        return not a
    
    @classmethod
    def _eval_reverse(cls, b):
        return not b


class NoneType:
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.UNDEFINED = super(NoneType, cls).__new__(cls)
        return cls.UNDEFINED

    def __repr__(self) -> str:
        return "UNDEFINED"
    
    def __bool__(self) -> bool:
        return False
    
    def __eq__(self, other: object) -> bool:
        return other is self

UNDEFINED = NoneType()
'''
Use this instead of None if you are using None as a value.
'''


import gc
from typing import List, TypeVar, Type

T = TypeVar('T')

def get_instances(cls: Type[T]) -> List[T]:
    """
    get all instances of a class.
    """
    return [
        object 
        for object in gc.get_objects()
        if isinstance(object, cls)
    ]


from typing import Callable
from colorama import Fore, Style
from ..debug import Log


def test(objects, *assert_functions, iterations=10):
    """
    test a list of objects with one or more given assert functions
    """

    PASSED = f"[Style: Bright, Color: Green]    PASSED[Style: Reset_All]"
    FAILED =   f"[Style: Bright, Color: Red]!!! FAILED[Style: Reset_All]"
    SUM_PASSED = f"[Style: Bright, Color: Green]PASSED[Style: Reset_All]"
    SUM_FAILED =   f"[Style: Bright, Color: Red]FAILED[Style: Reset_All]"

    if isinstance(objects, Callable):
        assert_functions = (*assert_functions, 
        objects)

    for func in assert_functions:
        func.takes_index = func.__code__.co_argcount > 1


    assert_function_strings = []

    max_raw_function_string_length = 0

    for func in assert_functions:
        assert_function_strings.append(
            f"[Style: Bright, Color: Magenta]{func.__name__.replace('_', ' ').title()}[Style: Reset_All]"
        )
        max_raw_function_string_length = max(max_raw_function_string_length, len(func.__name__))

    object_strings = []

    max_raw_object_string_length = 0

    for object_ in objects:
        if hasattr(object_, "name"):
            object_string = object_.name
        elif hasattr(object_, "__name__"):
            object_string = object_.__name__
        else:
            object_string = str(object_)
        
        max_raw_object_string_length = max(len(object_string), max_raw_object_string_length)
        
        object_string = f"[Style: Bright, Color: LightCyan_EX]{object_string}[Style: Reset_All]"
        
        object_strings.append(object_string)

    max_object_string_length = max(len(object_string) for object_string in object_strings)

    max_function_string_length = max(len(string) for string in assert_function_strings)
    
    summary_string = f"\n\n\n{' ' * int((max_raw_object_string_length + max_raw_function_string_length + 15) / 2)}--- Summary ---\n\n"

    for object_, object_string in zip(objects, object_strings):

        print("")
        
        for assert_function, assert_function_string in zip(assert_functions, assert_function_strings):
            fails = 0
            for i in range(iterations):
                return_string = ""
                try:
                    if assert_function.takes_index:
                        return_string = assert_function(object_, i)
                    else:
                        return_string = assert_function(object_)
                except AssertionError as e:
                    Log(f"{FAILED}: {object_string} - {assert_function_string} - Assertion failed: {e}")
                    fails += 1
                    continue
                except Exception as e:
                    Log(f"{FAILED}: {object_string} - {assert_function_string} - Exception thrown: {e}")
                    fails += 1
                    continue
                if not isinstance(return_string, str):
                    return_string = ""
                else:
                    return_string = f" with message: {return_string}"
                Log(f"{PASSED}: {object_string} - {assert_function_string} - Passed{return_string}")
            if fails == 0:
                summary_string += f"{SUM_PASSED}: {object_string: <{max_object_string_length}} - {assert_function_string: <{max_function_string_length}} - Passed all tests\n"
            else:
                summary_string += f"{SUM_FAILED}: {object_string: <{max_object_string_length}} - {assert_function_string: <{max_function_string_length}} - Failed {fails}/{iterations} tests\n"
    
    Log(summary_string)


from typing import Generic, TypeVar

T = TypeVar('T')

class Mutable(Generic[T]):
    '''
    mutable value to be passed by reference.
    implements dunder methods to point to the value.
    '''
    def __init__(self, value):
        self.value: T = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)
    
    def __get__(self, instance, owner):
        return self.value
    
    def __set__(self, instance, value):
        self.value = value
    
    def __delete__(self, instance):
        self.value = None
    
    def __call__(self, *args, **kwargs):
        return self.value()
    
    def __len__(self):
        return self.value.__len__()
    
    def __getitem__(self, key):
        return self.value.__getitem__(key)
    
    def __setitem__(self, key, value):
        self.value.__setitem__(key, value)
    
    def __delitem__(self, key):
        self.value.__delitem__(key)
    
    def __iter__(self):
        return self.value.__iter__()
    
    def __contains__(self, item):
        return self.value.__contains__(item)
    
    def __add__(self, other):
        if isinstance(other, Mutable):
            return self.value.__add__(other.value)
        else:
            return self.value.__add__(other)
        
    def __radd__(self, other):
        if isinstance(other, Mutable):
            return self.value.__radd__(other.value)
        else:
            return self.value.__radd__(other)
    
    def __iadd__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__iadd__(other.value)
            else:
                return self.value.__iadd__(other)
        except AttributeError:
            self.value = self.value.__add__(other)
            return self.value
    
    def __sub__(self, other):
        if isinstance(other, Mutable):
            return self.value.__sub__(other.value)
        else:
            return self.value.__sub__(other)
    
    def __rsub__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rsub__(other.value)
        else:
            return self.value.__rsub__(other)
    
    def __isub__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__isub__(other.value)
            else:
                return self.value.__isub__(other)
        except AttributeError:
            self.value = self.value.__sub__(other)
            return self.value

    def __mul__(self, other):
        if isinstance(other, Mutable):
            return self.value.__mul__(other.value)
        else:
            return self.value.__mul__(other)
    
    def __rmul__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rmul__(other.value)
        else:
            return self.value.__rmul__(other)
    
    def __imul__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__imul__(other.value)
            else:
                return self.value.__imul__(other)
        except AttributeError:
            self.value = self.value.__mul__(other)
            return self.value

    def __truediv__(self, other):
        if isinstance(other, Mutable):
            return self.value.__truediv__(other.value)
        else:
            return self.value.__truediv__(other)
    
    def __rtruediv__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rtruediv__(other.value)
        else:
            return self.value.__rtruediv__(other)
    
    def __itruediv__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__itruediv__(other.value)
            else:
                return self.value.__itruediv__(other)
        except AttributeError:
            self.value = self.value.__truediv__(other)
            return self.value

    def __floordiv__(self, other):
        if isinstance(other, Mutable):
            return self.value.__floordiv__(other.value)
        else:
            return self.value.__floordiv__(other)
    
    def __rfloordiv__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rfloordiv__(other.value)
        else:
            return self.value.__rfloordiv__(other)
    
    def __ifloordiv__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__ifloordiv__(other.value)
            else:
                return self.value.__ifloordiv__(other)
        except AttributeError:
            self.value = self.value.__floordiv__(other)
            return self.value

    def __mod__(self, other):
        if isinstance(other, Mutable):
            return self.value.__mod__(other.value)
        else:
            return self.value.__mod__(other)
    
    def __rmod__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rmod__(other.value)
        else:
            return self.value.__rmod__(other)
    
    def __imod__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__imod__(other.value)
            else:
                return self.value.__imod__(other)
        except AttributeError:
            self.value = self.value.__mod__(other)
            return self.value

    def __pow__(self, other):
        if isinstance(other, Mutable):
            return self.value.__pow__(other.value)
        else:
            return self.value.__pow__(other)
    
    def __rpow__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rpow__(other.value)
        else:
            return self.value.__rpow__(other)
    
    def __ipow__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__ipow__(other.value)
            else:
                return self.value.__ipow__(other)
        except AttributeError:
            self.value = self.value.__pow__(other)
            return self.value

    def __lt__(self, other):
        if isinstance(other, Mutable):
            return self.value.__lt__(other.value)
        else:
            return self.value.__lt__(other)
    
    def __le__(self, other):
        if isinstance(other, Mutable):
            return self.value.__le__(other.value)
        else:
            return self.value.__le__(other)
    
    def __eq__(self, other):
        if isinstance(other, Mutable):
            return self.value.__eq__(other.value)
        else:
            return self.value.__eq__(other)
    
    def __ne__(self, other):
        if isinstance(other, Mutable):
            return self.value.__ne__(other.value)
        else:
            return self.value.__ne__(other)
    
    def __ge__(self, other):
        if isinstance(other, Mutable):
            return self.value.__ge__(other.value)
        else:
            return self.value.__ge__(other)
    
    def __gt__(self, other):
        if isinstance(other, Mutable):
            return self.value.__gt__(other.value)
        else:
            return self.value.__gt__(other)
    
    def __and__(self, other):
        if isinstance(other, Mutable):
            return self.value.__and__(other.value)
        else:
            return self.value.__and__(other)
    
    def __rand__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rand__(other.value)
        else:
            return self.value.__rand__(other)
    
    def __iand__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__iand__(other.value)
            else:
                return self.value.__iand__(other)
        except AttributeError:
            self.value = self.value.__and__(other)
            return self.value

    def __or__(self, other):
        if isinstance(other, Mutable):
            return self.value.__or__(other.value)
        else:
            return self.value.__or__(other)
    
    def __ror__(self, other):
        if isinstance(other, Mutable):
            return self.value.__ror__(other.value)
        else:
            return self.value.__ror__(other)
    
    def __ior__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__ior__(other.value)
            else:
                return self.value.__ior__(other)
        except AttributeError:
            self.value = self.value.__or__(other)
            return self.value

    def __xor__(self, other):
        if isinstance(other, Mutable):
            return self.value.__xor__(other.value)
        else:
            return self.value.__xor__(other)
    
    def __rxor__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rxor__(other.value)
        else:
            return self.value.__rxor__(other)
    
    def __ixor__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__ixor__(other.value)
            else:
                return self.value.__ixor__(other)
        except AttributeError:
            self.value = self.value.__xor__(other)
            return self.value

    def __lshift__(self, other):
        if isinstance(other, Mutable):
            return self.value.__lshift__(other.value)
        else:
            return self.value.__lshift__(other)
    
    def __rlshift__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rlshift__(other.value)
        else:
            return self.value.__rlshift__(other)
    
    def __ilshift__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__ilshift__(other.value)
            else:
                return self.value.__ilshift__(other)
        except AttributeError:
            self.value = self.value.__lshift__(other)
            return self.value    

    def __rshift__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rshift__(other.value)
        else:
            return self.value.__rshift__(other)
    
    def __rrshift__(self, other):
        if isinstance(other, Mutable):
            return self.value.__rrshift__(other.value)
        else:
            return self.value.__rrshift__(other)
    
    def __irshift__(self, other):
        try:
            if isinstance(other, Mutable):
                return self.value.__irshift__(other.value)
            else:
                return self.value.__irshift__(other)
        except AttributeError:
            self.value = self.value.__rshift__(other)
            return self.value

    def __neg__(self):
        return self.value.__neg__()
    
    def __pos__(self):
        return self.value.__pos__()
    
    def __abs__(self):
        return self.value.__abs__()
    
    def __invert__(self):
        return self.value.__invert__()
    
    def __complex__(self):
        return self.value.__complex__()
    
    def __int__(self):
        return self.value.__int__()
    
    def __float__(self):
        return self.value.__float__()

    def __round__(self, n=0):
        return self.value.__round__(n)
    
    def __index__(self):
        return self.value.__index__()
    
    def __coerce__(self, other):
        return self.value.__coerce__(other)
    
    def __oct__(self):
        return self.value.__oct__()
    
    def __hex__(self):
        return self.value.__hex__()
    
    def __trunc__(self):
        return self.value.__trunc__()
    
    def __floor__(self):
        return self.value.__floor__()
    
    def __ceil__(self):
        return self.value.__ceil__()
    
    def __bool__(self):
        return self.value.__bool__()


import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="thelittlethings",
    version="0.9.1",
    author="Simon Kriele",
    author_email="kriele.simon@gmail.com",
    description="a library full of small utilities for you to use in your code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dots-git/thelittlethings",
    project_urls={
        "Bug Tracker": "https://github.com/dots-git/thelittlethings/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.0",
)